/**
 * @file geis_grail_backend.c
 * @brief GEIS grail client back end
 */

/*
 * Copyright 2011, 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License version 3, as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranties of 
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "geis_config.h"
#include "geis_grail_backend.h"

#include <float.h>
#include "geis_attr.h"
#include "geis_backend.h"
#include "geis_backend_protected.h"
#include "geis_bag.h"
#include "geis_event.h"
#include "geis_grail_token.h"
#include "geis_grail_window_grab.h"
#include "geis_group.h"
#include "geis_logging.h"
#include "geis_private.h"
#include "geis_touch.h"
#include "geis_ugsubscription_store.h"
#include "grail-types.h"
#include <math.h>
#include <string.h>
#include <utouch/frame_x11.h>
#include <utouch/grail.h>
#include <X11/extensions/sync.h>
#include <X11/extensions/XInput2.h>
#include <X11/Xlib.h>


static inline GeisSize
_min(GeisSize a, GeisSize b)
{
  return (a < b) ? a : b;
}


static inline GeisSize
_max(GeisSize a, GeisSize b)
{
  return (a > b) ? a : b;
}


struct _GeisSliceState
{
  unsigned int slice_id;
  uint64_t     timestamp;
  GeisFloat    angle;
  GeisFloat    position_x;
  GeisFloat    position_y;
  GeisFloat    radius;
};


/**
 * @addtogroup geis_backend_grail GEIS Grail Back End
 * @ingroup geis_backends
 *
 * A GEIS Back End that wraps a grail instance.
 *
 * @{
 */

/** The less opaque Grail Back End structure. */
struct GeisGrailBackend
{
  Geis                      geis;         /**< The parent GEIS instance. */
  Display                  *display;      /**< The X11 connection. */
  Window                    root_window;  /**< The X11 root window. */
  XSyncCounter              x11_timer;    /**< The X11 timer. */
  UFHandle                  frame;        /**< A utouch-frame instance. */
  UGHandle                  grail;        /**< A utouch-grail instance. */
  GeisBag                   devices;      /**< The list of known devices. */
  GeisGrailWindowGrabStore  window_grabs; /**< A collection of window inputs */
  GeisBag                   slice_states; /**< Gesture slice states */

  GeisGestureClass          drag_class;
  GeisGestureClass          pinch_class;
  GeisGestureClass          rotate_class;
  GeisGestureClass          tap_class;
  GeisGestureClass          touch_class;

  GeisBoolean               send_tentative_events;
  GeisBoolean               send_synchronous_events;
};


/**
 * Hashes an ID from a UFDevice.
 *
 * @todo this may need to be honed a little more.
 */
static inline GeisInteger
_grail_be_device_id_from_ufdevice(UFDevice ufdevice)
{
  return (intptr_t)ufdevice & 0xffff;
}


/**
 * Gets the UFDevice associated with an identified device hash.
 */
static UFDevice
_grail_be_ufdevice_from_device_id(GeisGrailBackend gbe, GeisInteger device_id)
{
  UFDevice ufdevice = NULL;
  for (GeisSize i = 0; i < geis_bag_count(gbe->devices); ++i)
  {
    UFDevice *d = geis_bag_at(gbe->devices, i);
    if (device_id == _grail_be_device_id_from_ufdevice(*d))
    {
      ufdevice = *d;
      break;
    }
  }
  return ufdevice;
}


/**
 * Extracts or calculates various gesture attributes.
 *
 * @param[in]  gbe         the grail back end instance
 * @param[in]  slice       the grail slice received
 * @param[in]  slice_state the current grail slice state
 * @param[in]  delta_t     the time since the previous frame
 * @param[out] frame       the gesture frame being created
 */
static void
_grail_be_extract_gesture_attrs(GeisGrailBackend        gbe,
                                UGSlice                 slice,
                                struct _GeisSliceState *slice_state,
                                GeisFloat               delta_t,
                                GeisFrame               frame)
{
  UGGestureTypeMask ugmask = grail_slice_get_recognized(slice);
  const UGTransform *C = grail_slice_get_cumulative_transform(slice);

  geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_GESTURE_NAME,
                                           GEIS_ATTR_TYPE_STRING,
                                           "n/a"));

  geis_frame_set_matrix(frame, &(*C)[0][0]);
  if (ugmask & UGGestureTypeDrag)
  {
    geis_frame_set_is_class(frame, gbe->drag_class);

    GeisFloat position_x = (*C)[0][2] + grail_slice_get_original_center_x(slice);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_POSITION_X,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &position_x));

    GeisFloat position_y = (*C)[1][2] + grail_slice_get_original_center_y(slice);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_POSITION_Y,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &position_y));

    GeisFloat delta_x = position_x - slice_state->position_x;
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_DELTA_X,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &delta_x));
    slice_state->position_x = position_x;

    GeisFloat delta_y = position_y - slice_state->position_y;
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_DELTA_Y,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &delta_y));
    slice_state->position_y = position_y;

    if (delta_t > 0.0f)
    {
      GeisFloat x_velocity = delta_x / delta_t;
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_VELOCITY_X,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &x_velocity));

      GeisFloat y_velocity = delta_y / delta_t;
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_VELOCITY_Y,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &y_velocity));
    }
  }
  if (ugmask & UGGestureTypePinch)
  {
    geis_frame_set_is_class(frame, gbe->pinch_class);

    GeisFloat ca = (*C)[0][0];
    GeisFloat cb = (*C)[0][1];
    GeisFloat r = sqrtf(ca*ca + cb*cb);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_RADIUS,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &r));

    GeisFloat dr = r - slice_state->radius;
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_RADIUS_DELTA,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &dr));
    slice_state->radius = r;

    if (delta_t > 0.0f)
    {
      GeisFloat vr = (delta_t > 0.0f) ? dr / delta_t : 0.0f;
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_RADIAL_VELOCITY,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &vr));
    }
  }
  if (ugmask & UGGestureTypeRotate)
  {
    geis_frame_set_is_class(frame, gbe->rotate_class);

    GeisFloat centre_x = grail_slice_get_center_of_rotation_x(slice);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_CENTROID_X,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &centre_x));

    GeisFloat centre_y = grail_slice_get_center_of_rotation_y(slice);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_CENTROID_Y,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &centre_y));
    GeisFloat angle = atan2((*C)[0][1], (*C)[0][0]);
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_ANGLE_DELTA,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &angle));

    GeisFloat da = angle - slice_state->angle;
    geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_ANGLE,
                                             GEIS_ATTR_TYPE_FLOAT,
                                             &da));
    slice_state->angle = angle;

    if (delta_t > 0.0f)
    {
      GeisFloat va = (delta_t > 0.0f) ? da / delta_t : 0.0f;
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_ANGULAR_VELOCITY,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &va));
    }
  }
  if (ugmask & UGGestureTypeTap)
  {
    geis_frame_set_is_class(frame, gbe->tap_class);
  }
  if (ugmask & UGGestureTypeTouch)
  {
    geis_frame_set_is_class(frame, gbe->touch_class);
  }
}


/**
 * Gets the previous slice state (if any) for a grail slice.
 *
 * @param[in] gbe    the grail back end
 * @param[in] slice  a grail slice
 *
 * @returns a pointer to a grail slice state or NULL if there is no previous
 * grail slice state.
 */
struct _GeisSliceState *
_grail_be_slice_state_from_ugslice(GeisGrailBackend gbe,
                                   UGSlice          slice)
{
  struct _GeisSliceState *slice_state = NULL;
  for (GeisSize i = 0; i < geis_bag_count(gbe->slice_states); ++i)
  {
    struct _GeisSliceState *s = geis_bag_at(gbe->slice_states, i);
    if (s->slice_id == grail_slice_get_id(slice))
    {
      slice_state = s;
      break;
    }
  }
  return slice_state;
}


/**
 * Constructs a new slice state for a grail slise.
 *
 * @param[in] gbe        the grail back end
 * @param[in] slice      a grail slice
 *
 * @returns a pointer to a grail slice state or NULL on failure.
 */
struct _GeisSliceState *
_grail_be_slice_state_new(GeisGrailBackend gbe,
                          UGSlice          slice)
{
  struct _GeisSliceState new_slice_state = {
    .slice_id = grail_slice_get_id(slice),
    .timestamp = 0, 
    .angle = 0.0f,
    .position_x = 0.0f,
    .position_y = 0.0f,
    .radius = 0.0f
  };
  geis_bag_append(gbe->slice_states, &new_slice_state);
  return _grail_be_slice_state_from_ugslice(gbe, slice);
}


/**
 * Creates a geis event from a grail slice.
 *
 * @param[in] gbe          the grail back end
 * @param[in] slice        the grail slice
 * @param[in] clice_state  the slice state (may be NULL)
 *
 * @returns a new geis event or NULL if no valid event can be created for the
 * slice.
 */
static GeisEvent
_grail_be_geis_event_from_ugslice(GeisGrailBackend        gbe,
                                  UGSlice                 slice,
                                  struct _GeisSliceState *slice_state)
{
  GeisEvent geis_event = NULL;

  /* Retrieve the slice's recognized gestures. */
  UGGestureTypeMask ugmask = grail_slice_get_recognized(slice);

  switch (grail_slice_get_state(slice))
  {
    case UGGestureStateBegin:
      if (ugmask)
        geis_event = geis_event_new(GEIS_EVENT_GESTURE_BEGIN);
      else if (gbe->send_tentative_events)
        geis_event = geis_event_new(GEIS_EVENT_TENTATIVE_BEGIN);
      break;
    case UGGestureStateUpdate:
      if (ugmask)
        if (slice_state)
          geis_event = geis_event_new(GEIS_EVENT_GESTURE_UPDATE);
        else
          geis_event = geis_event_new(GEIS_EVENT_GESTURE_BEGIN);
      else if (gbe->send_tentative_events)
        geis_event = geis_event_new(GEIS_EVENT_TENTATIVE_UPDATE);
      break;
    case UGGestureStateEnd:
      if (slice_state)
        geis_event = geis_event_new(GEIS_EVENT_GESTURE_END);
      else if (gbe->send_tentative_events)
        geis_event = geis_event_new(GEIS_EVENT_TENTATIVE_END);
      break;
  }

  return geis_event;
}


static void
_grail_be_grail_fd_callback(int                             fd GEIS_UNUSED,
                            GeisBackendMultiplexorActivity  ev GEIS_UNUSED,
                            void                           *ctx)
{
  GeisGrailBackend gbe = (GeisGrailBackend)ctx;
  UGEvent event;
  while (grail_get_event(gbe->grail, &event) == UGStatusSuccess)
  {
    /* Extract the grail slice from the grail event. */
    if (UGEventTypeSlice == grail_event_get_type(event))
    {
      UGSlice slice;
      UGStatus ugstatus = grail_event_get_property(event,
                                                   UGEventPropertySlice,
                                                   &slice);
      if (ugstatus != UGStatusSuccess)
      {
        geis_error("could not retrieve slice from grail event");
        goto next_event;
      }

      /* Get any previous slice state for the grail slice. */
      struct _GeisSliceState *slice_state = NULL;
      slice_state = _grail_be_slice_state_from_ugslice(gbe, slice);

      /* Translate the grail slice into a geis event, maybe. */
      GeisEvent geis_event = _grail_be_geis_event_from_ugslice(gbe,
                                                               slice,
                                                               slice_state);
      if (!geis_event)
      {
        goto next_event;
      }

      /* A gesture needs to start out with an initial state. */
      if (GEIS_EVENT_GESTURE_BEGIN == geis_event_type(geis_event))
      {
        slice_state = _grail_be_slice_state_new(gbe, slice);
        if (!slice_state)
        {
          geis_error("logic error");
          geis_event_delete(geis_event);
          goto next_event;
        }
      }

      /* Updated any saved gesture state. */
      uint64_t timestamp = grail_event_get_time(event);
      GeisFloat time_delta = 0.0f;
      if (slice_state)
      {
        uint64_t dt = timestamp - slice_state->timestamp;
        /* Skip synchronous events if not requested to send them. */
        if (GEIS_EVENT_GESTURE_UPDATE == geis_event_type(geis_event)
          && !gbe->send_synchronous_events
          && dt <= 0)
        {
          geis_event_delete(geis_event);
          goto next_event;
        }
        time_delta = dt;
        slice_state->timestamp = timestamp;
      }

      GeisGroupSet  groupset = geis_groupset_new();
      GeisAttr      group_attr = geis_attr_new(GEIS_EVENT_ATTRIBUTE_GROUPSET,
                                               GEIS_ATTR_TYPE_POINTER,
                                               groupset);
      geis_attr_set_destructor(group_attr,
                               (GeisAttrDestructor)geis_groupset_delete);
      GeisGroup group = geis_group_new(1);
      geis_groupset_insert(groupset, group);

      GeisTouchSet  touchset = geis_touchset_new();
      GeisAttr      touch_attr = geis_attr_new(GEIS_EVENT_ATTRIBUTE_TOUCHSET,
                                               GEIS_ATTR_TYPE_POINTER,
                                               touchset);
      geis_attr_set_destructor(touch_attr,
                               (GeisAttrDestructor)geis_touchset_delete);

      GeisFrame frame = geis_frame_new(grail_slice_get_id(slice));
      geis_group_insert_frame(group, frame);

      GeisInteger ts = timestamp;
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_TIMESTAMP,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &ts));

      UFFrame ufframe = grail_slice_get_frame(slice);
      UFDevice ufdevice = frame_frame_get_device(ufframe);
      GeisInteger geis_device_id = _grail_be_device_id_from_ufdevice(ufdevice);
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_DEVICE_ID,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &geis_device_id));

      GeisBoolean is_touchscreen = GEIS_FALSE;
      GeisDevice geis_device = geis_get_device(gbe->geis, geis_device_id);
      if (!geis_device)
      {
        geis_warning("unrecognized device %d reported", geis_device_id);
      }
      else
      {
        GeisAttr attr = geis_device_attr_by_name(geis_device,
                                   GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH);
        is_touchscreen = (attr && geis_attr_value_to_boolean(attr));
      }

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_ROOT_WINDOW_ID,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &gbe->root_window));

      GeisInteger window = frame_x11_get_window_id(frame_frame_get_window_id(ufframe));
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_EVENT_WINDOW_ID,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &window));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_CHILD_WINDOW_ID,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &window));

      GeisInteger num_touches = grail_slice_get_num_touches(slice);
      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_TOUCHES,
                                               GEIS_ATTR_TYPE_INTEGER,
                                               &num_touches));

      GeisFloat focus_x = 0.0f;
      GeisFloat focus_y = 0.0f;
      if (is_touchscreen)
      {
        const UGTransform *C = grail_slice_get_cumulative_transform(slice);
        focus_x = (*C)[0][2] + grail_slice_get_original_center_x(slice);
        focus_y = (*C)[1][2] + grail_slice_get_original_center_y(slice);
      }

      GeisFloat bboxMinX, bboxMinY = FLT_MAX;
      GeisFloat bboxMaxX, bboxMaxY = 0.0f;
      for (GeisInteger t = 0; t < num_touches; ++t)
      {
        GeisTouch touch = geis_touch_new(t);
        if (!touch)
        {
          geis_error("can not create GEIS touch object");
          continue;
        }

        UFTouchId touch_id;
        ugstatus = grail_slice_get_touch_id(slice, t, &touch_id);
        GeisInteger touch_slot = touch_id;
        geis_touch_add_attr(touch, geis_attr_new(GEIS_TOUCH_ATTRIBUTE_ID,
                                                 GEIS_ATTR_TYPE_INTEGER,
                                                 &touch_slot));

        UFTouch uftouch;
        UFStatus ufstatus;
        ufstatus = frame_frame_get_touch_by_id(ufframe, touch_id, &uftouch);
        if (ufstatus != UFStatusSuccess)
        {
          geis_error("can not retrieve touch %ld from slice frame", (long)touch_id);
          continue;
        }

        if (!is_touchscreen)
        {
          focus_x = frame_touch_get_window_x(uftouch);
          focus_y = frame_touch_get_window_y(uftouch);
        }

        GeisFloat fval = frame_touch_get_device_x(uftouch);
        bboxMinX = fmin(bboxMinX, fval);
        bboxMaxX = fmax(bboxMaxX, fval);
        geis_touch_add_attr(touch, geis_attr_new(GEIS_TOUCH_ATTRIBUTE_X,
                                                 GEIS_ATTR_TYPE_FLOAT,
                                                 &fval));

        fval = frame_touch_get_device_y(uftouch);
        bboxMinY = fmin(bboxMinY, fval);
        bboxMaxY = fmax(bboxMaxY, fval);
        geis_touch_add_attr(touch, geis_attr_new(GEIS_TOUCH_ATTRIBUTE_Y,
                                                 GEIS_ATTR_TYPE_FLOAT,
                                                 &fval));

        geis_touchset_insert(touchset, touch);
        geis_frame_add_touchid(frame, geis_touch_id(touch));
      }

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_FOCUS_X,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &focus_x));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_FOCUS_Y,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &focus_y));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_BOUNDINGBOX_X1,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &bboxMinX));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_BOUNDINGBOX_Y1,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &bboxMinY));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_BOUNDINGBOX_X2,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &bboxMaxX));

      geis_frame_add_attr(frame, geis_attr_new(GEIS_GESTURE_ATTRIBUTE_BOUNDINGBOX_Y2,
                                               GEIS_ATTR_TYPE_FLOAT,
                                               &bboxMaxY));

      _grail_be_extract_gesture_attrs(gbe,
                                      slice, slice_state,
                                      time_delta,
                                      frame);
      geis_event_add_attr(geis_event, group_attr);
      geis_event_add_attr(geis_event, touch_attr);
      geis_post_event(gbe->geis, geis_event);

      /* Destroy any saved state after the a gesture has ended. */
      if (UGGestureStateEnd == grail_slice_get_state(slice))
      {
        for (GeisSize i = 0; i < geis_bag_count(gbe->slice_states); ++i)
        {
          if (slice_state == geis_bag_at(gbe->slice_states, i))
          {
            geis_bag_remove(gbe->slice_states, i);
            break;
          }
        }
      }
    }

next_event:
    grail_event_unref(event);
  }

  /* Set a timeouton the X server, if necessary. */
  uint64_t timeout = grail_next_timeout(gbe->grail);
  if (timeout)
  {
    XSyncWaitCondition condition;

    condition.trigger.counter = gbe->x11_timer;
    condition.trigger.value_type = XSyncAbsolute;
    XSyncIntsToValue(&condition.trigger.wait_value,
                     timeout & 0x00000000ffffffff,
                     timeout & 0xffffffff00000000);
    condition.trigger.test_type = XSyncPositiveComparison;
    XSyncIntToValue(&condition.event_threshold, 0);
    if (!XSyncAwait(gbe->display, &condition, 1))
    {
      geis_warning("failed to set wait trigger");
    }
    XFlush(gbe->display);
  }
}


/**
 * Reports an X11 device to the front end as a GEIS device.
 */
static void
_geis_grail_add_device(GeisGrailBackend gbe, UFDevice frame_device)
{
  UFStatus status;

  geis_bag_append(gbe->devices, &frame_device);
  GeisInteger device_id = _grail_be_device_id_from_ufdevice(frame_device);

  char *device_name = NULL;
  status = frame_device_get_property(frame_device,
                                     UFDevicePropertyName,
                                     &device_name);
  if (status != UFStatusSuccess)
  {
    geis_error("failed to get 'name' property from device");
    goto final_exit;
  }

  GeisDevice geis_device = geis_device_new(device_name, device_id);
  if (!geis_device)
  {
    geis_error("failed to create GEIS device");
    goto final_exit;
  }

  int ival;
  status = frame_device_get_property(frame_device,
                                     UFDevicePropertyMaxTouches,
                                     &ival);
  if (status != UFStatusSuccess)
  {
    geis_warning("failed to get 'touches' property from device '%s'",
                 device_name);
  }
  else
  {
    GeisAttr attr = geis_attr_new(GEIS_DEVICE_ATTRIBUTE_TOUCHES,
                                  GEIS_ATTR_TYPE_INTEGER,
                                  &ival);
    geis_device_add_attr(geis_device, attr);
  }

  status = frame_device_get_property(frame_device,
                                     UFDevicePropertyDirect,
                                     &ival);
  if (status != UFStatusSuccess)
  {
    geis_warning("failed to get 'direct' property from device '%s'",
                 device_name);
  }
  else
  {
    GeisAttr attr = geis_attr_new(GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH,
                                  GEIS_ATTR_TYPE_BOOLEAN,
                                  &ival);
    geis_device_add_attr(geis_device, attr);
  }

  status = frame_device_get_property(frame_device,
                                     UFDevicePropertyIndependent,
                                     &ival);
  if (status != UFStatusSuccess)
  {
    geis_warning("failed to get 'independent' property from device '%s'",
                 device_name);
  }
  else
  {
    GeisAttr attr = geis_attr_new(GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH,
                                  GEIS_ATTR_TYPE_BOOLEAN,
                                  &ival);
    geis_device_add_attr(geis_device, attr);
  }

  /* Report the device as a filterable entity. */
  static struct GeisFilterableAttribute attrs[] = {
    { GEIS_DEVICE_ATTRIBUTE_NAME,              GEIS_ATTR_TYPE_STRING,  0, NULL },
    { GEIS_DEVICE_ATTRIBUTE_ID,                GEIS_ATTR_TYPE_INTEGER, 0, NULL },
    { GEIS_DEVICE_ATTRIBUTE_TOUCHES,           GEIS_ATTR_TYPE_INTEGER, 0, NULL },
    { GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH,      GEIS_ATTR_TYPE_BOOLEAN, 0, NULL },
    { GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH, GEIS_ATTR_TYPE_BOOLEAN, 0, NULL },
  };
  static GeisSize attr_count = sizeof(attrs)
                             / sizeof(struct GeisFilterableAttribute);

  geis_register_device(gbe->geis, geis_device, attr_count, attrs);
  goto final_exit;

final_exit:
  return;
}


static void
_geis_grail_remove_device(GeisGrailBackend gbe, UFDevice frame_device)
{
  for (GeisSize i = 0; i < geis_bag_count(gbe->devices); ++i)
  {
    UFDevice *device = geis_bag_at(gbe->devices, i);
    GeisInteger device_id = _grail_be_device_id_from_ufdevice(*device);
    if (device_id == _grail_be_device_id_from_ufdevice(frame_device))
    {
      GeisDevice geis_device = geis_get_device(gbe->geis, device_id);
      if (!geis_device)
      {
        geis_warning("unrecognized device %d has been removed", device_id);
      }
      else
      {
        geis_unregister_device(gbe->geis, geis_device);
      }
      break;
    }
  }
}


/**
 * Processes all the utouch-frame events until the queue is empty.
 *
 * @param[in] gbe        the grail back end
 */
static void
_geis_be_flush_frame_events(GeisGrailBackend gbe)
{
  UFEvent frame_event;
  while (frame_get_event(gbe->frame, &frame_event) == UFStatusSuccess)
  {
    UFEventType event_type = frame_event_get_type(frame_event);
    if (event_type == UFEventTypeDeviceAdded)
    {
      UFDevice frame_device;
      UFStatus status = frame_event_get_property(frame_event,
                                                 UFEventPropertyDevice,
                                                 &frame_device);
      if (status != UFStatusSuccess)
      {
        geis_warning("can not get device from device-added frame event");
      }
      else
      {
        _geis_grail_add_device(gbe, frame_device);
      }
    }
    else if (event_type == UFEventTypeDeviceRemoved)
    {
      UFDevice frame_device;
      UFStatus status = frame_event_get_property(frame_event,
                                                 UFEventPropertyDevice,
                                                 &frame_device);
      if (status != UFStatusSuccess)
      {
        geis_warning("can not get device from device-removed frame event");
      }
      else
      {
        _geis_grail_remove_device(gbe, frame_device);
      }
    }
    grail_process_frame_event(gbe->grail, frame_event);
    frame_event_unref(frame_event);
  }
}


/**
 * Callback invoked when utouch-frame has events in the queue.
 *
 * @param[in] fd   The utouch-frame file descriptor (not used).
 * @param[in] ev   The type of FD event (not used).
 * @param[in] ctx  The context data (the GEIS grail back end).
 */
static void
_grail_be_frame_fd_callback(int                             fd GEIS_UNUSED,
                            GeisBackendMultiplexorActivity  ev GEIS_UNUSED,
                            void                           *ctx)
{
  GeisGrailBackend gbe = (GeisGrailBackend)ctx;
  _geis_be_flush_frame_events(gbe);
}


/**
 * Pushes X11 events into the utouch-frame filter.
 *
 * @param[in] fd   The X11 file descriptor (not used).
 * @param[in] ev   The type of FD event (not used).
 * @param[in] ctx  The context data (the GEIS grail back end).
 */
static void
_x11_fd_callback(int                             fd GEIS_UNUSED,
                 GeisBackendMultiplexorActivity  ev GEIS_UNUSED,
                 void                           *ctx)
{
  GeisGrailBackend gbe = (GeisGrailBackend)ctx;
  XEvent event;
  while (XPending(gbe->display))
  {
    XNextEvent(gbe->display, &event);

    /* Non-Generic events should be timer events. */
    if (event.type != GenericEvent)
    {
      if (gbe->x11_timer)
      {
        _geis_be_flush_frame_events(gbe);

        XSyncValue sync_time;
        if (XSyncQueryCounter(gbe->display, gbe->x11_timer, &sync_time) != True)
        {
          geis_warning("failed to get X server time");
        }
        uint64_t grail_time = (uint64_t)XSyncValueHigh32(sync_time) << 32
                                      | XSyncValueLow32(sync_time);
        grail_update_time(gbe->grail, grail_time);
      }
      continue;
    }

    XGenericEventCookie *xcookie = &event.xcookie;
    if (!XGetEventData(gbe->display, xcookie))
    {
      geis_warning("failed to get X generic event data");
      continue;
    }

    UFStatus frame_status = frame_x11_process_event(gbe->frame, xcookie);
    if (frame_status != UFStatusSuccess)
    {
      geis_warning("failed to inject X11 event");
    }

    XFreeEventData(gbe->display, xcookie);
  }
}


/**
 * Checks if the X11 server supports the appropriate version of the XInput
 * extension.
 *
 * @param[in] gbe  The grail back end
 *
 * @returns GEIS_TRUE if the X11 server provides the XInput extension of a
 * sufficiently recent vintage, GEIS_FALSE otherwise.
 */
static GeisBoolean
_geis_grail_x11_has_xi2(GeisGrailBackend gbe)
{
  GeisBoolean has_xi2 = GEIS_FALSE;

  /* check if the XInput extension is available at all */
  int opcode;
  int event;
  int error;
  if (!XQueryExtension(gbe->display, "XInputExtension", &opcode, &event, &error))
  {
    geis_error("XInput extension is not available");
    goto final_exit;
  }

  /* check if it's the right version */
  int major = 2;
  int minor = 2;
  if (XIQueryVersion(gbe->display, &major, &minor) == BadRequest)
  {
    geis_error("XI2 is unavailable, X Server supports only %d.%d", major, minor);
    goto final_exit;
  }
  has_xi2 = GEIS_TRUE;

final_exit:
  return has_xi2;
}


/**
 * Gets a synchronization timer from the X11 server (if one is available).
 *
 * @param[in] gbe  The grail back end
 *
 * @returns a valid syncronization timer from the X11 server is an appropriate
 * one is available, otherwise returns a value < 0.
 */
static XSyncCounter
_geis_grail_x11_get_timer(GeisGrailBackend gbe)
{
  XSyncCounter x11_timer = -1;

  /* check for the presence of the XSync extension */
  int event;
  int error;
  if (XSyncQueryExtension(gbe->display, &event, &error) != True)
  {
    geis_warning("XSync extension is not available");
    goto final_exit;
  }

  /* initialize the XSync extension */
  int major;
  int minor;
  if (XSyncInitialize(gbe->display, &major, &minor) != True)
  {
    geis_warning("failed to initialize XSync extension");
    goto final_exit;
  }

  /* find the server-time counter */
  int num_system_counters;
  XSyncSystemCounter *counters;
  counters = XSyncListSystemCounters(gbe->display, &num_system_counters);
  for (int i = 0; i < num_system_counters; ++i)
  {
    if (0 == strcmp(counters[i].name, "SERVERTIME"))
    {
      x11_timer = counters[i].counter;
      break;
    }
  }
  XSyncFreeSystemCounterList(counters);

final_exit:
  return x11_timer;
}


/**
 * Connects to the X11 server and sets up processing on X11 events.
 *
 * @param[in] gbe  The grail back end
 *
 * @returns GEIS_TRUE if a conneciton to X11 server has been successfully
 * extablished and it provides the minimum required functionality,
 * GEIS_FALSE otherwise.
 */
static GeisBoolean
_geis_grail_open_x11_connection(GeisGrailBackend gbe)
{
  GeisBoolean success = GEIS_FALSE;
  gbe->display = XOpenDisplay(NULL);
  if (!gbe->display)
  {
    geis_error("error connecting to X server");
    goto final_exit;
  }

  /* verify all the necessary extensions are available */
  if (!_geis_grail_x11_has_xi2(gbe))
  {
    goto final_exit;
  }

  gbe->root_window = DefaultRootWindow(gbe->display);
  gbe->x11_timer = _geis_grail_x11_get_timer(gbe);

  /* install an X11 event callback */
  geis_multiplex_fd(gbe->geis,
                    ConnectionNumber(gbe->display),
                    GEIS_BE_MX_READ_AVAILABLE,
                    _x11_fd_callback,
                    gbe);

  success = GEIS_TRUE;

final_exit:
  return success;
}


/**
 * Asks the X server for device added or removed events.
 */
static void
_geis_grail_subscribe_x11_device_events(GeisGrailBackend gbe)
{
  XIEventMask mask = {
    .deviceid = XIAllDevices,
    .mask_len = XIMaskLen(XI_LASTEVENT),
    .mask     = calloc(XIMaskLen(XI_LASTEVENT), sizeof(char))
  };
  XISetMask(mask.mask, XI_HierarchyChanged);
  Status status = XISelectEvents(gbe->display, gbe->root_window, &mask, 1);
  if (status != Success)
  {
    geis_error("error %d selecting device-changed events on X server", status);
  }
  free(mask.mask);
}


/**
 * Reports available gesture classes to the front end.
 */
static void
_geis_grail_report_classes(GeisGrailBackend gbe)
{
  static struct GeisFilterableAttribute attrs[] = {
    {
      GEIS_CLASS_ATTRIBUTE_NAME,
      GEIS_ATTR_TYPE_STRING,
      0,
      NULL
    },
    {
      GEIS_CLASS_ATTRIBUTE_ID,
      GEIS_ATTR_TYPE_INTEGER,
      0,
      NULL
    },
    {
      GEIS_GESTURE_ATTRIBUTE_TOUCHES,
      GEIS_ATTR_TYPE_INTEGER,
      0,
      NULL
    }
  };
  GeisSize attr_count = sizeof(attrs)
                      / sizeof(struct GeisFilterableAttribute);

  gbe->drag_class = geis_gesture_class_new(GEIS_GESTURE_DRAG, GRAIL_TYPE_DRAG1);
  geis_register_gesture_class(gbe->geis, gbe->drag_class, attr_count, attrs);

  gbe->pinch_class = geis_gesture_class_new(GEIS_GESTURE_PINCH, GRAIL_TYPE_PINCH1);
  geis_register_gesture_class(gbe->geis, gbe->pinch_class, attr_count, attrs);

  gbe->rotate_class = geis_gesture_class_new(GEIS_GESTURE_ROTATE, GRAIL_TYPE_ROTATE1);
  geis_register_gesture_class(gbe->geis, gbe->rotate_class, attr_count, attrs);

  gbe->tap_class = geis_gesture_class_new(GEIS_GESTURE_TAP, GRAIL_TYPE_TAP1);
  geis_register_gesture_class(gbe->geis, gbe->tap_class, attr_count, attrs);

  gbe->touch_class = geis_gesture_class_new(GEIS_GESTURE_TOUCH, GRAIL_TYPE_TOUCH1);
  geis_register_gesture_class(gbe->geis, gbe->touch_class, attr_count, attrs);
}


/**
 * Reports available regions to the front end.
 */
static void
_geis_grail_report_regions(GeisGrailBackend gbe)
{
  static struct GeisFilterableAttribute attrs[] = {
    { 
      GEIS_REGION_ATTRIBUTE_WINDOWID,
      GEIS_ATTR_TYPE_INTEGER,
      0,
      NULL
    },
  };
  static const GeisSize attr_count = sizeof(attrs)
                                   / sizeof(struct GeisFilterableAttribute);

  geis_register_region(gbe->geis, NULL, attr_count, attrs);
}


/**
 * Reports a successful initialization to the front end.
 */
static void
_geis_grail_report_init_complete(GeisGrailBackend gbe)
{
  geis_post_event(gbe->geis, geis_event_new(GEIS_EVENT_INIT_COMPLETE));
}


/**
 * Constructs a Grail back end.
 *
 * @param[in] mem
 * @param[in] geis
 */
static void
_geis_grail_backend_construct(void *mem, Geis geis)
{
  GeisGrailBackend gbe = (GeisGrailBackend)mem;
  gbe->geis = geis;

  if (_geis_grail_open_x11_connection(gbe))
  {
    _geis_grail_subscribe_x11_device_events(gbe);

    UFStatus frame_status = frame_x11_new(gbe->display, &gbe->frame);
    if (frame_status != UFStatusSuccess)
    {
      geis_error("failed to create utouch frame instance");
      goto unwind_x11;
    }
    geis_multiplex_fd(gbe->geis,
                      frame_get_fd(gbe->frame),
                      GEIS_BE_MX_READ_AVAILABLE,
                      _grail_be_frame_fd_callback,
                      gbe);

    UGStatus grail_status = grail_new(gbe->frame, &gbe->grail);
    if (grail_status != UGStatusSuccess)
    {
      geis_error("failed to create utouch grail instance");
      goto unwind_frame;
    }
    geis_multiplex_fd(gbe->geis,
                      grail_get_fd(gbe->grail),
                      GEIS_BE_MX_READ_AVAILABLE,
                      _grail_be_grail_fd_callback,
                      gbe);

    gbe->devices = geis_bag_new(sizeof(UFDevice), 2, 2.0f);
    if (!gbe->devices)
    {
      geis_error("failed to create UFDevices store");
      goto unwind_grail;
    }

    gbe->window_grabs = geis_grail_window_grab_store_new(gbe->display);
    if (!gbe->window_grabs)
    {
      geis_error("failed to create window grabs store");
      goto unwind_devices;
    }

    gbe->slice_states = geis_bag_new(sizeof(struct _GeisSliceState), 2, 1.5f);
    if (!gbe->slice_states)
    {
      geis_error("failed to create slice times store");
      goto unwind_grabs;
    }

    geis_get_configuration(gbe->geis,
                           GEIS_CONFIG_SEND_TENTATIVE_EVENTS,
                           &gbe->send_tentative_events);
    geis_get_configuration(gbe->geis,
                           GEIS_CONFIG_SEND_SYNCHRONOS_EVENTS,
                           &gbe->send_synchronous_events);

    _geis_grail_report_classes(gbe);
    _geis_grail_report_regions(gbe);
    _geis_grail_report_init_complete(gbe);
    goto final_exit;

unwind_grabs:
    geis_grail_window_grab_store_delete(gbe->window_grabs);
unwind_devices:
    geis_bag_delete(gbe->devices);
unwind_grail:
    geis_demultiplex_fd(gbe->geis, grail_get_fd(gbe->grail));
    grail_delete_v3(gbe->grail);
unwind_frame:
    geis_demultiplex_fd(gbe->geis, frame_get_fd(gbe->frame));
    frame_x11_delete(gbe->frame);
unwind_x11:
    geis_demultiplex_fd(gbe->geis, ConnectionNumber(gbe->display));
    XCloseDisplay(gbe->display);
  }

final_exit:
  return;
}


/**
 * Deconstructs a Grail back end.
 *
 * @param[in] be  A %GeisGrailBackend.
 */
static void 
_geis_grail_backend_finalize(GeisBackend be)
{
  GeisGrailBackend gbe = (GeisGrailBackend)be;
  geis_bag_delete(gbe->slice_states);
  geis_grail_window_grab_store_delete(gbe->window_grabs);
  geis_bag_delete(gbe->devices);
  geis_demultiplex_fd(gbe->geis, grail_get_fd(gbe->grail));
  grail_delete_v3(gbe->grail);
  geis_demultiplex_fd(gbe->geis, frame_get_fd(gbe->frame));
  frame_x11_delete(gbe->frame);
  geis_demultiplex_fd(gbe->geis, ConnectionNumber(gbe->display));
  XCloseDisplay(gbe->display);
}


static struct GeisBackendVtable gbe_vtbl = {
  _geis_grail_backend_construct,
  _geis_grail_backend_finalize,
  geis_grail_token_new
};


static GeisStatus
_geis_grail_filter_gestures(GeisGrailBackend gbe,
                            GeisFilter       filter,
                            GeisSubscription sub,
                            UGSubscription   ugsub)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;
  GeisSubscriptionFlags sub_flags = geis_subscription_flags(sub);

  /* default mask is ALL gesture (for now) */
  UGGestureTypeMask default_ugmask = UGGestureTypeDrag
                                   | UGGestureTypePinch
                                   | UGGestureTypeRotate
                                   | UGGestureTypeTap
                                   | UGGestureTypeTouch;
  UGGestureTypeMask ugmask = 0;
  unsigned int start_touches = 0;
  unsigned int min_touches = 1;
  unsigned int max_touches = 5;

  for (GeisSize i = 0; i < geis_filter_term_count(filter); ++i)
  {
    GeisFilterTerm term = geis_filter_term(filter, i);
    if (GEIS_FILTER_CLASS == geis_filter_term_facility(term))
    {
      GeisAttr            attr      = geis_filter_term_attr(term);
      GeisString          name      = geis_attr_name(attr);
      GeisFilterOperation operation = geis_filter_term_operation(term);

      if (0 == strcmp(name, GEIS_CLASS_ATTRIBUTE_NAME)
          && operation == GEIS_FILTER_OP_EQ)
      {
        GeisString class_name = geis_attr_value_to_string(attr);
        if (0 == strcmp(class_name, geis_gesture_class_name(gbe->drag_class)))
          ugmask |= UGGestureTypeDrag;
        else if (0 == strcmp(class_name, geis_gesture_class_name(gbe->pinch_class)))
          ugmask |= UGGestureTypePinch;
        else if (0 == strcmp(class_name, geis_gesture_class_name(gbe->rotate_class)))
          ugmask |= UGGestureTypeRotate;
        else if (0 == strcmp(class_name, geis_gesture_class_name(gbe->tap_class)))
          ugmask |= UGGestureTypeTap;
        else if (0 == strcmp(class_name, geis_gesture_class_name(gbe->touch_class)))
          ugmask |= UGGestureTypeTouch;
      }
      else if (0 == strcmp(name, GEIS_CLASS_ATTRIBUTE_ID)
          && operation == GEIS_FILTER_OP_EQ)
      {
        GeisInteger class_id = geis_attr_value_to_integer(attr);
        if (class_id == geis_gesture_class_id(gbe->drag_class))
          ugmask |= UGGestureTypeDrag;
        else if (class_id == geis_gesture_class_id(gbe->pinch_class))
          ugmask |= UGGestureTypePinch;
        else if (class_id == geis_gesture_class_id(gbe->rotate_class))
          ugmask |= UGGestureTypeRotate;
        else if (class_id == geis_gesture_class_id(gbe->tap_class))
          ugmask |= UGGestureTypeTap;
        else if (class_id == geis_gesture_class_id(gbe->touch_class))
          ugmask |= UGGestureTypeTouch;
      }
      else if (0 == strcmp(name, GEIS_GESTURE_ATTRIBUTE_TOUCHES))
      {
        /*
         * Starting touches: filter terms are ANDed: multiple TOUCHES
         * terms are ANDed using the MAX operator.
         */
        unsigned int val = geis_attr_value_to_integer(attr);
        switch (operation)
        {
          case GEIS_FILTER_OP_EQ:
            start_touches = _max(start_touches, val);
            min_touches = _max(min_touches, start_touches);
            max_touches = _min(max_touches, start_touches);
            break;
          case GEIS_FILTER_OP_NE:
            geis_error("unsupported comparison");
            break;
          case GEIS_FILTER_OP_GE:
            start_touches = _max(start_touches, val);
            min_touches = _max(min_touches, start_touches);
            max_touches = _min(max_touches, 5);
            break;
          case GEIS_FILTER_OP_GT:
            start_touches = _max(start_touches, val+1);
            min_touches = _max(min_touches, start_touches);
            max_touches = _min(max_touches, 5);
            break;
          case GEIS_FILTER_OP_LE:
            start_touches = _max(start_touches, val);
            min_touches = _max(min_touches, 0);
            max_touches = _min(max_touches, start_touches);
            break;
          case GEIS_FILTER_OP_LT:
            start_touches = _max(start_touches, val-1);
            min_touches = _max(min_touches, 0);
            max_touches = _min(max_touches, start_touches);
            break;
        }
      }
    }
  }

  if (sub_flags & GEIS_SUBSCRIPTION_CONT)
  {
    min_touches = 1;
    max_touches = start_touches;
  }
  if (!ugmask)
    ugmask = default_ugmask;

  geis_debug("mask=0x%08x start=%u min=%u max=%u",
             ugmask, start_touches, min_touches, max_touches);

  UGStatus ugstatus;
  ugstatus = grail_subscription_set_property(ugsub,
                                             UGSubscriptionPropertyMask,
                                             &ugmask);
  if (ugstatus != UGStatusSuccess)
  {
    geis_error("failed to set UGSubscription mask");
    goto final_exit;
  }

  ugstatus = grail_subscription_set_property(ugsub,
                                             UGSubscriptionPropertyTouchesStart,
                                             &start_touches);
  if (ugstatus != UGStatusSuccess)
  {
    geis_error("failed to set UGSubscription start touches");
    goto final_exit;
  }

  ugstatus = grail_subscription_set_property(ugsub,
                                             UGSubscriptionPropertyTouchesMinimum,
                                             &min_touches);
  if (ugstatus != UGStatusSuccess)
  {
    geis_error("failed to set UGSubscription min touches");
    goto final_exit;
  }

  ugstatus = grail_subscription_set_property(ugsub,
                                             UGSubscriptionPropertyTouchesMaximum,
                                             &max_touches);
  if (ugstatus != UGStatusSuccess)
  {
    geis_error("failed to set UGSubscription max touches");
    goto final_exit;
  }

  status = GEIS_STATUS_SUCCESS;

final_exit:
  return status;
}


/**
 * Activates a subscription for a (device, region).
 */
static GeisStatus
_geis_grail_activate_for_device_region(GeisGrailBackend gbe,
                                       GeisFilter       filter,
                                       GeisDevice       device,
                                       GeisInteger      window_id,
                                       GeisSubscription subscription)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;
  GeisInteger device_id = geis_device_id(device);
  UFDevice ufdevice = _grail_be_ufdevice_from_device_id(gbe, device_id);
  GeisUGSubscriptionStore ugstore = geis_subscription_pdata(subscription);
  UGSubscription ugsub = geis_ugsubscription_get_ugsubscription(ugstore,
                                                                filter,
                                                                ufdevice,
                                                                window_id);
  if (!ugsub)
  {
    geis_error("can not retrieve UGSubscription for (device, window)");
    goto final_exit;
  }

  UGStatus ugstatus = UGStatusErrorGeneric;

  GeisBoolean geis_use_atomic_gestures = GEIS_FALSE;
  geis_get_configuration(gbe->geis,
                         GEIS_CONFIG_ATOMIC_GESTURES,
                         &geis_use_atomic_gestures);
  int grail_use_atomic_gestures = (geis_use_atomic_gestures == GEIS_TRUE);
  ugstatus = grail_subscription_set_property(ugsub,
                                             UGSubscriptionPropertyAtomicGestures,
                                             &grail_use_atomic_gestures);

  status = _geis_grail_filter_gestures(gbe, filter, subscription, ugsub);

  UGGestureTypeMask ugmask;
  ugstatus = grail_subscription_get_property(ugsub,
                                             UGSubscriptionPropertyMask,
                                             &ugmask);
  if (ugstatus != UGStatusSuccess)
  {
    geis_error("failed to get UGSubscription mask");
    goto final_exit;
  }

  geis_debug("subscription='%s' filter='%s' device=%d '%s' window=0x%08x "
             "ugsub=%p atomic=%d",
             geis_subscription_name(subscription),
             geis_filter_name(filter),
             device_id, geis_device_name(device),
             window_id,
             (void *)ugsub,
             grail_use_atomic_gestures);

  geis_grail_window_grab_store_grab(gbe->window_grabs, window_id);
  ugstatus = grail_subscription_activate(gbe->grail, ugsub);
  if (ugstatus != UGStatusSuccess)
  {
    status = GEIS_STATUS_UNKNOWN_ERROR;
    geis_error("failed to activate UGSubscription");
  }

final_exit:
  return status;
}


/**
 * Activates a subscription for a list of devices.
 */
static GeisStatus
_geis_grail_activate_for_devices(GeisGrailBackend gbe,
                                 GeisFilter       filter,
                                 GeisDeviceBag    device_bag,
                                 GeisSubscription subscription)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;
  for (GeisSize d = 0; d < geis_device_bag_count(device_bag); ++d)
  {
    GeisDevice device = geis_device_bag_device(device_bag, d);
    GeisBoolean has_region_term = GEIS_FALSE;
    for (GeisSize i = 0; i < geis_filter_term_count(filter); ++i)
    {
      GeisFilterTerm term = geis_filter_term(filter, i);
      if (GEIS_FILTER_REGION == geis_filter_term_facility(term))
      {
        GeisInteger         window_id = 0;
        GeisAttr            attr      = geis_filter_term_attr(term);
        GeisString          name      = geis_attr_name(attr);
        GeisFilterOperation operation = geis_filter_term_operation(term);

        if (0 == strcmp(name, GEIS_REGION_ATTRIBUTE_WINDOWID)
          && operation == GEIS_FILTER_OP_EQ)
        {
          window_id = geis_attr_value_to_integer(attr);
          has_region_term = GEIS_TRUE;
          status = _geis_grail_activate_for_device_region(gbe,
                                                          filter,
                                                          device,
                                                          window_id,
                                                          subscription);
        }
        else
        {
          geis_warning("unhandled region filter term");
        }
      }
    }

    /* If no region terms were found, use the root window. */
    if (!has_region_term)
    {
      status = _geis_grail_activate_for_device_region(gbe,
                                                      filter,
                                                      device,
                                                      gbe->root_window,
                                                      subscription);
    }
  }
  return status;
}


/*
 * Activates a GEIS subscription on the back end.
 */
GeisStatus
geis_grail_backend_activate_subscription(GeisGrailBackend gbe,
                                         GeisSubscription subscription)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;

  GeisUGSubscriptionStore ugstore = geis_subscription_pdata(subscription);
  if (ugstore == NULL)
  {
    ugstore = geis_ugsubscription_store_new();
    if (!ugstore)
    {
      geis_error("error creating grail subscription store");
      goto final_exit;
    }
    geis_subscription_set_pdata(subscription, ugstore);
  }

  for (GeisFilterIterator it = geis_subscription_filter_begin(subscription);
       it != geis_subscription_filter_end(subscription);
       it = geis_subscription_filter_next(subscription, it))
  {
    GeisFilter filter = *it;
    GeisDeviceBag selected_devices = geis_device_bag_new();
    GeisSelectResult device_matches = geis_select_devices(gbe->geis,
                                                          filter,
                                                          selected_devices);
    if (device_matches == GEIS_SELECT_RESULT_ALL)
    {
      geis_debug("filter %p '%s' matches ALL devices",
                 (void *)filter,
                 geis_filter_name(filter));

      /* Special case: no devices succeeds, in case one gets added later. */
      if (geis_device_bag_count(geis_devices(gbe->geis)) == 0)
        status = GEIS_STATUS_SUCCESS;
      else
        status = _geis_grail_activate_for_devices(gbe,
                                                  filter,
                                                  geis_devices(gbe->geis),
                                                  subscription);
    }
    else if (device_matches == GEIS_SELECT_RESULT_SOME)
    {
      geis_debug("filter %p '%s' matches %zu devices",
                 (void *)filter,
                 geis_filter_name(filter),
                 geis_device_bag_count(selected_devices));
      status = _geis_grail_activate_for_devices(gbe,
                                                filter,
                                                selected_devices,
                                                subscription);
    }
    else
    {
      geis_debug("filter %p '%s' matches NO devices",
                 (void *)filter,
                 geis_filter_name(filter));
      status = GEIS_STATUS_SUCCESS;
    }
    geis_device_bag_delete(selected_devices);
  }

final_exit:
  return status;
}


/*
 * Deactivates a GEIS subscription on the back end.
 *
 * @todo replace this stub
 */
GeisStatus
geis_grail_backend_deactivate_subscription(GeisGrailBackend gbe,
                                           GeisSubscription subscription)
{
  GeisStatus status = GEIS_STATUS_UNKNOWN_ERROR;

  GeisUGSubscriptionStore ugstore = geis_subscription_pdata(subscription);
  if (ugstore)
  {
    for (GeisSize i = 0; i < geis_ugsubscription_count(ugstore); ++i)
    {
      UGSubscription ugsub = geis_ugsubscription_get_ugsubscription_at(ugstore,
                                                                       i);
      grail_subscription_deactivate(gbe->grail, ugsub);
    }
  }

  return status;
}


/**
 * Registers the back end with the GEIS back end registry.
 */
static void __attribute__((constructor))
_register_grail_backend(void)
{
  geis_register_backend(GEIS_INIT_UTOUCH_GRAIL_BACKEND,
                        sizeof(struct GeisGrailBackend),
                        &gbe_vtbl);
}


/** A dummy routine to force linkage of this module without dlopening it */
void
geis_include_grail_backend(void)
{
}

/** @} */
